# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import abc
import sys
import typing as ta

from ..lite.maysyncs import Maywaitable
from ..lite.maysyncs import make_maysync
from ..lite.maysyncs import maysync
from ..lite.timeouts import TimeoutLike
from .asyncs import AbstractAsyncSubprocesses
from .base import BaseSubprocesses
from .run import SubprocessRun
from .run import SubprocessRunOutput
from .sync import AbstractSubprocesses


##


class AbstractMaysyncSubprocesses(BaseSubprocesses, abc.ABC):
    @abc.abstractmethod
    def run_(self, run: SubprocessRun) -> Maywaitable[SubprocessRunOutput]:
        raise NotImplementedError

    def run(
            self,
            *cmd: str,
            input: ta.Any = None,  # noqa
            timeout: TimeoutLike = None,
            check: bool = False,
            capture_output: ta.Optional[bool] = None,
            **kwargs: ta.Any,
    ) -> Maywaitable[SubprocessRunOutput]:
        return self.run_(SubprocessRun(
            cmd=cmd,
            input=input,
            timeout=timeout,
            check=check,
            capture_output=capture_output,
            kwargs=kwargs,
        ))

    #

    @abc.abstractmethod
    def check_call(
            self,
            *cmd: str,
            stdout: ta.Any = sys.stderr,
            **kwargs: ta.Any,
    ) -> Maywaitable[None]:
        raise NotImplementedError

    @abc.abstractmethod
    def check_output(
            self,
            *cmd: str,
            **kwargs: ta.Any,
    ) -> Maywaitable[bytes]:
        raise NotImplementedError

    #

    @maysync
    async def check_output_str(
            self,
            *cmd: str,
            **kwargs: ta.Any,
    ) -> str:
        return (await self.check_output(*cmd, **kwargs).a()).decode().strip()

    #

    @maysync
    async def try_call(
            self,
            *cmd: str,
            **kwargs: ta.Any,
    ) -> bool:
        if isinstance(await self.async_try_fn(self.check_call(*cmd, **kwargs).a), Exception):
            return False
        else:
            return True

    @maysync
    async def try_output(
            self,
            *cmd: str,
            **kwargs: ta.Any,
    ) -> ta.Optional[bytes]:
        if isinstance(ret := await self.async_try_fn(self.check_output(*cmd, **kwargs).a), Exception):
            return None
        else:
            return ret

    @maysync
    async def try_output_str(
            self,
            *cmd: str,
            **kwargs: ta.Any,
    ) -> ta.Optional[str]:
        if (ret := await self.try_output(*cmd, **kwargs).a()) is None:
            return None
        else:
            return ret.decode().strip()


##


class MaysyncSubprocesses(AbstractMaysyncSubprocesses):
    def __init__(
            self,
            subprocesses: AbstractSubprocesses,
            async_subprocesses: AbstractAsyncSubprocesses,
    ) -> None:
        super().__init__()

        self._subprocesses = subprocesses
        self._async_subprocesses = async_subprocesses

    #

    def run_(self, run: SubprocessRun) -> Maywaitable[SubprocessRunOutput]:
        return make_maysync(
            self._subprocesses.run,
            self._async_subprocesses.run,
        )(run)

    #

    def check_call(
            self,
            *cmd: str,
            stdout: ta.Any = sys.stderr,
            **kwargs: ta.Any,
    ) -> Maywaitable[None]:
        return make_maysync(
            self._subprocesses.check_call,
            self._async_subprocesses.check_call,
        )(*cmd, stdout=stdout, **kwargs)

    def check_output(
            self,
            *cmd: str,
            **kwargs: ta.Any,
    ) -> Maywaitable[bytes]:
        return make_maysync(
            self._subprocesses.check_output,
            self._async_subprocesses.check_output,
        )(*cmd, **kwargs)



import typing as ta

from .base import AnyArrayKeyword
from .base import AnyKeyword
from .base import BooleanKeyword
from .base import BooleanOrKeywordsKeyword
from .base import Keyword
from .base import Keywords
from .base import KeywordsArrayKeyword
from .base import KeywordsKeyword
from .base import NumberKeyword
from .base import StrKeyword
from .base import StrOrStrArrayKeyword
from .base import StrToKeywordsKeyword
from .base import UnknownKeyword


##


def render_keyword(kw: Keyword) -> dict[str, ta.Any]:
    if isinstance(kw, AnyKeyword):
        return {kw.tag: kw.v}

    elif isinstance(kw, AnyArrayKeyword):
        return {kw.tag: kw.vs}

    elif isinstance(kw, BooleanKeyword):
        return {kw.tag: kw.b}

    elif isinstance(kw, BooleanOrKeywordsKeyword):
        if isinstance(kw.bk, bool):
            return {kw.tag: kw.bk}
        else:
            return {kw.tag: render_keywords(kw.bk)}

    elif isinstance(kw, NumberKeyword):
        return {kw.tag: kw.n}

    elif isinstance(kw, StrKeyword):
        return {kw.tag: kw.s}

    elif isinstance(kw, StrOrStrArrayKeyword):
        if isinstance(kw.ss, str):
            return {kw.tag: kw.ss}
        else:
            return {kw.tag: list(kw.ss)}

    elif isinstance(kw, KeywordsKeyword):
        return {kw.tag: render_keywords(kw.kw)}

    elif isinstance(kw, KeywordsArrayKeyword):
        return {kw.tag: [render_keywords(c) for c in kw.kws]}

    elif isinstance(kw, StrToKeywordsKeyword):
        return {kw.tag: {k: render_keywords(v) for k, v in kw.m.items()}}

    elif isinstance(kw, UnknownKeyword):
        return {kw.tag: kw.value}

    else:
        raise TypeError(kw)


def render_keywords(kws: Keywords) -> dict[str, ta.Any]:
    dct: dict[str, ta.Any] = {}
    for kw in kws.lst:
        kwd = render_keyword(kw)
        [(tag, val)] = kwd.items()
        if tag in dct:
            raise KeyError(tag)
        dct[tag] = val
    return dct



import builtins
import contextlib
import functools
import importlib.util
import types
import typing as ta

from ..lazyglobals import LazyGlobals


##


class _ProxyInit:
    class NamePackage(ta.NamedTuple):
        name: str
        package: str

    class _Import(ta.NamedTuple):
        pkg: str
        attr: str | None

    def __init__(
            self,
            lazy_globals: LazyGlobals,
            name_package: NamePackage,
    ) -> None:
        super().__init__()

        self._lazy_globals = lazy_globals
        self._name_package = name_package

        self._imps_by_attr: dict[str, _ProxyInit._Import] = {}
        self._mods_by_pkgs: dict[str, ta.Any] = {}

    @property
    def name_package(self) -> NamePackage:
        return self._name_package

    def add(
            self,
            package: str,
            attrs: ta.Iterable[tuple[str | None, str]],
    ) -> None:
        for imp_attr, as_attr in attrs:
            if imp_attr is None:
                self._imps_by_attr[as_attr] = self._Import(package, None)
                self._lazy_globals.set_fn(as_attr, functools.partial(self.get, as_attr))

            else:
                self._imps_by_attr[as_attr] = self._Import(package, imp_attr)
                self._lazy_globals.set_fn(as_attr, functools.partial(self.get, as_attr))

    def _import_module(self, name: str) -> ta.Any:
        return importlib.import_module(name, package=self._name_package.package)

    def get(self, attr: str) -> ta.Any:
        try:
            imp = self._imps_by_attr[attr]
        except KeyError:
            raise AttributeError(attr)  # noqa

        val: ta.Any

        if imp.attr is None:
            val = self._import_module(imp.pkg)

        else:
            try:
                mod = self._mods_by_pkgs[imp.pkg]
            except KeyError:
                mod = self._import_module(imp.pkg)
                self._mods_by_pkgs[imp.pkg] = mod

            val = getattr(mod, imp.attr)

        return val


def proxy_init(
        init_globals: ta.MutableMapping[str, ta.Any],
        package: str,
        attrs: ta.Iterable[str | tuple[str | None, str | None] | None] | None = None,
) -> None:
    if isinstance(attrs, str):
        raise TypeError(attrs)

    if attrs is None:
        attrs = [None]

    whole_attr = package.split('.')[-1]
    al: list[tuple[str | None, str]] = []
    for attr in attrs:
        if attr is None:
            al.append((None, whole_attr))

        elif isinstance(attr, str):
            al.append((attr, attr))

        elif isinstance(attr, tuple):
            imp_attr, as_attr = attr
            if as_attr is None:
                if imp_attr is None:
                    as_attr = whole_attr
                else:
                    as_attr = imp_attr
            al.append((imp_attr, as_attr))

        else:
            raise TypeError(attr)

    #

    init_name_package = _ProxyInit.NamePackage(
        init_globals['__name__'],
        init_globals['__package__'],
    )

    pi: _ProxyInit
    try:
        pi = init_globals['__proxy_init__']

    except KeyError:
        pi = _ProxyInit(
            LazyGlobals.install(init_globals),
            init_name_package,
        )
        init_globals['__proxy_init__'] = pi

    else:
        if pi.name_package != init_name_package:
            raise Exception(f'Wrong init name: {pi.name_package=} != {init_name_package=}')

    pi.add(package, al)


##


class AutoProxyInitError(Exception):
    pass


class AutoProxyInitErrors:
    def __new__(cls, *args, **kwargs):  # noqa
        raise TypeError

    class HookError(AutoProxyInitError):
        pass

    class AttrError(AutoProxyInitError):
        def __init__(self, module: str | None, name: str) -> None:
            super().__init__()

            self.module = module
            self.name = name

        def __repr__(self) -> str:
            return f'{self.__class__.__qualname__}(module={self.module!r}, name={self.name!r})'

    class ImportError(AutoProxyInitError):  # noqa
        def __init__(self, module: str, from_list: ta.Sequence[str] | None) -> None:
            super().__init__()

            self.module = module
            self.from_list = from_list

        def __repr__(self) -> str:
            return f'{self.__class__.__qualname__}(module={self.module!r}, from_list={self.from_list!r})'

    class ImportStarForbiddenError(ImportError):
        pass

    class UnproxiedImportForbiddenError(ImportError):
        pass

    class UnreferencedImportsError(AutoProxyInitError):
        def __init__(self, unreferenced: ta.Mapping[str, ta.Sequence[str | None]]) -> None:
            super().__init__()

            self.unreferenced = unreferenced

        def __repr__(self) -> str:
            return f'{self.__class__.__qualname__}(unreferenced={self.unreferenced!r})'


class _AutoProxyInitCapture:
    class ModuleSpec(ta.NamedTuple):
        name: str
        level: int

        def __str__(self) -> str:
            return f'{"." * self.level}{self.name}'

        def __repr__(self) -> str:
            return repr(str(self))

    def __init__(self) -> None:
        super().__init__()

        self._modules_by_spec: dict[_AutoProxyInitCapture.ModuleSpec, _AutoProxyInitCapture._Module] = {}
        self._modules_by_module_obj: dict[types.ModuleType, _AutoProxyInitCapture._Module] = {}

        self._attrs: dict[_AutoProxyInitCapture._ModuleAttr, tuple[_AutoProxyInitCapture._Module, str]] = {}

    #

    class _ModuleAttr:
        def __init__(
                self,
                module: '_AutoProxyInitCapture._Module',
                name: str,
        ) -> None:
            super().__init__()

            self.__module = module
            self.__name = name

        def __repr__(self) -> str:
            return f'<{self.__class__.__name__}: {f"{self.__module.spec}:{self.__name}"!r}>'

    class _Module:
        def __init__(
                self,
                spec: '_AutoProxyInitCapture.ModuleSpec',
                *,
                getattr_handler: ta.Callable[['_AutoProxyInitCapture._Module', str], ta.Any] | None = None,
        ) -> None:
            super().__init__()

            self.spec = spec

            self.module_obj = types.ModuleType(f'<{self.__class__.__qualname__}: {spec!r}>')
            if getattr_handler is not None:
                self.module_obj.__getattr__ = functools.partial(getattr_handler, self)  # type: ignore[method-assign]  # noqa
            self.initial_module_dict = dict(self.module_obj.__dict__)

            self.contents: dict[str, _AutoProxyInitCapture._ModuleAttr | types.ModuleType] = {}
            self.imported_whole = False

        def __repr__(self) -> str:
            return f'{self.__class__.__name__}({self.spec!r})'

    def _get_or_make_module(self, spec: ModuleSpec) -> _Module:
        try:
            return self._modules_by_spec[spec]
        except KeyError:
            pass

        module = self._Module(
            spec,
            getattr_handler=self._handle_module_getattr,
        )
        self._modules_by_spec[spec] = module
        self._modules_by_module_obj[module.module_obj] = module
        return module

    def _handle_module_getattr(self, module: _Module, attr: str) -> ta.Any:
        if attr in module.contents:
            raise AutoProxyInitErrors.AttrError(str(module.spec), attr)

        v: _AutoProxyInitCapture._ModuleAttr | types.ModuleType
        if not module.spec.name:
            if not module.spec.level:
                raise AutoProxyInitError
            cs = _AutoProxyInitCapture.ModuleSpec(attr, module.spec.level)
            cm = self._get_or_make_module(cs)
            cm.imported_whole = True
            v = cm.module_obj

        else:
            ma = _AutoProxyInitCapture._ModuleAttr(module, attr)
            self._attrs[ma] = (module, attr)
            v = ma

        module.contents[attr] = v
        setattr(module.module_obj, attr, v)
        return v

    def _handle_import(
            self,
            module: _Module,
            *,
            from_list: ta.Sequence[str] | None,
    ) -> None:
        if from_list is None:
            if module.spec.level or not module.spec.name:
                raise AutoProxyInitError

            module.imported_whole = True

        else:
            for attr in from_list:
                if attr == '*':
                    raise AutoProxyInitErrors.ImportStarForbiddenError(str(module.spec), from_list)

                x = getattr(module.module_obj, attr)

                bad = False
                if x is not module.contents.get(attr):
                    bad = True
                if isinstance(x, _AutoProxyInitCapture._ModuleAttr):
                    if self._attrs[x] != (module, attr):
                        bad = True
                elif isinstance(x, types.ModuleType):
                    if x not in self._modules_by_module_obj:
                        bad = True
                else:
                    bad = True
                if bad:
                    raise AutoProxyInitErrors.AttrError(str(module.spec), attr)

    #

    _MOD_SELF_ATTR: ta.ClassVar[str] = '__auto_proxy_init_capture__'

    def _intercept_import(
            self,
            name: str,
            *,
            globals: ta.Mapping[str, ta.Any] | None = None,  # noqa
            from_list: ta.Sequence[str] | None = None,
            level: int = 0,
    ) -> types.ModuleType | None:
        if not (
                globals is not None and
                globals.get(self._MOD_SELF_ATTR) is self
        ):
            return None

        spec = _AutoProxyInitCapture.ModuleSpec(name, level)
        module = self._get_or_make_module(spec)

        self._handle_import(
            module,
            from_list=from_list,
        )

        return module.module_obj

    @contextlib.contextmanager
    def hook_context(
            self,
            init_globals: ta.MutableMapping[str, ta.Any],  # noqa
            *,
            forbid_unproxied_imports: bool = False,
    ) -> ta.Iterator[None]:
        if self._MOD_SELF_ATTR in init_globals:
            raise AutoProxyInitErrors.HookError

        old_import = builtins.__import__

        def new_import(
                name,
                globals=None,  # noqa
                locals=None,  # noqa
                fromlist=None,
                level=0,
        ):
            if (im := self._intercept_import(
                    name,
                    globals=globals,
                    from_list=fromlist,
                    level=level,
            )) is not None:
                return im

            if forbid_unproxied_imports:
                raise AutoProxyInitErrors.UnproxiedImportForbiddenError(
                    str(_AutoProxyInitCapture.ModuleSpec(name, level)),
                    fromlist,
                )

            return old_import(
                name,
                globals=globals,
                locals=locals,
                fromlist=fromlist,
                level=level,
            )

        #

        init_globals[self._MOD_SELF_ATTR] = self
        builtins.__import__ = new_import

        try:
            yield

        finally:
            if not (
                    init_globals[self._MOD_SELF_ATTR] is self and
                    builtins.__import__ is new_import
            ):
                raise AutoProxyInitErrors.HookError

            del init_globals[self._MOD_SELF_ATTR]
            builtins.__import__ = old_import

    #

    def verify_state(
            self,
            init_globals: ta.MutableMapping[str, ta.Any],  # noqa
    ) -> None:
        for m in self._modules_by_spec.values():
            for a, o in m.module_obj.__dict__.items():
                try:
                    i = m.initial_module_dict[a]

                except KeyError:
                    if o is not m.contents[a]:
                        raise AutoProxyInitErrors.AttrError(str(m.spec), a) from None

                else:
                    if o != i:
                        raise AutoProxyInitErrors.AttrError(str(m.spec), a)

    #

    class ProxyInit(ta.NamedTuple):
        package: str
        attrs: ta.Sequence[tuple[str | None, str]]

    class BuiltProxyInits(ta.NamedTuple):
        proxy_inits: ta.Sequence['_AutoProxyInitCapture.ProxyInit']
        unreferenced: ta.Mapping[str, ta.Sequence[str | None]] | None

    def build_proxy_inits(
            self,
            init_globals: ta.MutableMapping[str, ta.Any],  # noqa
            *,
            collect_unreferenced: bool = False,
    ) -> BuiltProxyInits:
        dct: dict[_AutoProxyInitCapture._Module, list[tuple[str | None, str]]] = {}

        rem_whole_mods: set[_AutoProxyInitCapture._Module] = set()
        rem_mod_attrs: set[_AutoProxyInitCapture._ModuleAttr] = set()
        if collect_unreferenced:
            rem_whole_mods.update([m for m in self._modules_by_spec.values() if m.imported_whole])
            rem_mod_attrs.update(self._attrs)

        for attr, obj in init_globals.items():
            if isinstance(obj, _AutoProxyInitCapture._ModuleAttr):
                try:
                    m, a = self._attrs[obj]
                except KeyError:
                    raise AutoProxyInitErrors.AttrError(None, attr) from None
                dct.setdefault(m, []).append((a, attr))
                rem_mod_attrs.discard(obj)

            elif isinstance(obj, _AutoProxyInitCapture._Module):
                raise AutoProxyInitErrors.AttrError(None, attr) from None

            elif isinstance(obj, types.ModuleType):
                try:
                    m = self._modules_by_module_obj[obj]
                except KeyError:
                    continue
                if not m.imported_whole:
                    raise RuntimeError(f'AutoProxyInit module {m.spec!r} not imported_whole')
                dct.setdefault(m, []).append((None, attr))
                rem_whole_mods.discard(m)

        lst: list[_AutoProxyInitCapture.ProxyInit] = []
        for m, ts in dct.items():
            if not m.spec.name:
                if not m.spec.level:
                    raise AutoProxyInitError
                for imp_attr, as_attr in ts:
                    if not imp_attr:
                        raise RuntimeError
                    lst.append(_AutoProxyInitCapture.ProxyInit(
                        '.' * m.spec.level + imp_attr,
                        [(None, as_attr)],
                    ))

            else:
                lst.append(_AutoProxyInitCapture.ProxyInit(
                    str(m.spec),
                    ts,
                ))

        unreferenced: dict[str, list[str | None]] | None = None
        if collect_unreferenced and (rem_whole_mods or rem_mod_attrs):
            unreferenced = {}
            for m in rem_whole_mods:
                unreferenced.setdefault(str(m.spec), []).append(None)
            for ma in rem_mod_attrs:
                m, a = self._attrs[ma]
                unreferenced.setdefault(str(m.spec), []).append(a)

        return _AutoProxyInitCapture.BuiltProxyInits(
            lst,
            unreferenced,
        )


@contextlib.contextmanager
def auto_proxy_init(
        init_globals: ta.MutableMapping[str, ta.Any],
        *,
        disable: bool = False,
        eager: bool = False,

        unreferenced_callback: ta.Callable[[ta.Mapping[str, ta.Sequence[str | None]]], None] | None = None,
        raise_unreferenced: bool = False,
) -> ta.Iterator[None]:
    """
    This is a bit extreme - use sparingly. It relies on an interpreter-global import lock, but much of the ecosystem
    implicitly does anyway. It further relies on temporarily patching `__builtins__.__import__`, but could be switched
    to use any number of other import hooks.
    """

    if disable:
        yield
        return

    cap = _AutoProxyInitCapture()

    with cap.hook_context(init_globals):
        yield

    cap.verify_state(init_globals)

    blt = cap.build_proxy_inits(
        init_globals,
        collect_unreferenced=unreferenced_callback is not None or raise_unreferenced,
    )

    if blt.unreferenced:
        if unreferenced_callback:
            unreferenced_callback(blt.unreferenced)
        if raise_unreferenced:
            raise AutoProxyInitErrors.UnreferencedImportsError(blt.unreferenced)

    for pi in blt.proxy_inits:
        for _, a in pi.attrs:
            del init_globals[a]

        proxy_init(
            init_globals,
            pi.package,
            pi.attrs,
        )

    if eager:
        lg = LazyGlobals.install(init_globals)

        for pi in blt.proxy_inits:
            for _, a in pi.attrs:
                lg.get(a)



import typing as ta

from ..lite.maysyncs import MaysyncGenerator
from ..lite.maysyncs import Maywaitable
from ..lite.maysyncs import make_maysync as _make_maysync
from ..lite.maysyncs import maysync as _maysync
from ..lite.maysyncs import maysync_fn as _maysync_fn
from ..lite.maysyncs import maysync_generator_fn as _maysync_generator_fn
from .functions import as_async


T = ta.TypeVar('T')
O = ta.TypeVar('O')
I = ta.TypeVar('I')

P = ta.ParamSpec('P')

MaysyncFn: ta.TypeAlias = ta.Callable[P, Maywaitable[T]]
MaysyncGeneratorFn: ta.TypeAlias = ta.Callable[P, MaysyncGenerator[O, I]]


##


def make_maysync(
        s: ta.Callable[P, T],
        a: ta.Callable[P, ta.Awaitable[T]],
) -> MaysyncFn[P, T]:
    return ta.cast('MaysyncFn[P, T]', _make_maysync(
        s,
        a,
    ))


def make_maysync_from_sync(
        s: ta.Callable[P, T],
        a: ta.Callable[P, ta.Awaitable[T]] | None = None,
) -> MaysyncFn[P, T]:
    return ta.cast('MaysyncFn[P, T]', _make_maysync(
        s,
        a if a is not None else as_async(s),
    ))


##


def maysync_fn(
        m: ta.Callable[P, ta.Awaitable[T]],
) -> MaysyncFn[P, T]:
    return ta.cast('MaysyncFn[P, T]', _maysync_fn(m))


def maysync_generator_fn(
        m: ta.Callable[P, ta.AsyncGenerator[O, I]],
) -> MaysyncGeneratorFn[P, O, I]:
    return ta.cast('MaysyncGeneratorFn[P, O, I]', _maysync_generator_fn(m))


@ta.overload
def maysync(
        m: ta.Callable[P, ta.Awaitable[T]],
) -> MaysyncFn[P, T]:
    ...


@ta.overload
def maysync(
        m: ta.Callable[P, ta.AsyncGenerator[O, I]],
) -> MaysyncGeneratorFn[P, O, I]:
    ...


def maysync(m):
    return _maysync(m)



import typing as ta


T = ta.TypeVar('T')
P = ta.ParamSpec('P')


##


async def async_list(
        fn: ta.Callable[P, ta.AsyncIterator[T]],
        *args: P.args,
        **kwargs: P.kwargs,
) -> list[T]:
    """Simply eagerly reads the full contents of a function call returning an async iterator."""

    return [v async for v in fn(*args, **kwargs)]


##


def sync_await(
        fn: ta.Callable[P, ta.Awaitable[T]],
        *args: P.args,
        **kwargs: P.kwargs,
) -> T:
    """
    Allows for the synchronous execution of async functions which will never actually *externally* await anything. These
    functions are allowed to await any number of other functions - including contextmanagers and generators - so long as
    nothing ever actually 'leaks' out of the function, presumably to an event loop.
    """

    ret = missing = object()

    async def gate():
        nonlocal ret

        ret = await fn(*args, **kwargs)

    cr = gate()
    try:
        try:
            cr.send(None)
        except StopIteration:
            pass

        if ret is missing or cr.cr_await is not None or cr.cr_running:
            raise TypeError('Not terminated')

    finally:
        cr.close()

    return ta.cast(T, ret)


def sync_async_list(
        fn: ta.Callable[P, ta.AsyncIterator[T]],
        *args: P.args,
        **kwargs: P.kwargs,
) -> list[T]:
    """
    Uses `sync_await` to synchronously read the full contents of a function call returning an async iterator, given that
    the function never externally awaits anything.
    """

    lst: list[T] | None = None

    async def inner():
        nonlocal lst

        lst = [v async for v in fn(*args, **kwargs)]

    sync_await(inner)

    if not isinstance(lst, list):
        raise TypeError(lst)

    return lst



# ruff: noqa: UP006 UP007 UP043 UP045 UP046 UP047
# @omlish-lite
"""
A system for writing a python function once which can then be effectively used in both sync and async contexts -
including IO, under any (or no) event loop.

Where an 'async fn' returns an 'awaitable', a 'maysync fn' returns a 'maywaitable', which is an object with two
nullary methods:

 - `def s()` - to be called in sync contexts
 - `async def a()` - to be called in async and maysync contexts

For example, a maysync function `m_inc_int(x: int) -> int` would be used as such:

 - `assert m_inc_int(5).s() == 6` in sync contexts
 - `assert await m_inc_int(5).a() == 6` in async and maysync contexts

Maysync fns may be constructed in two ways: either using `make_maysync`, providing an equivalent pair of sync and async
functions, or using the `@maysync` decorator to wrap a 'maysync flavored' async function. 'Maysync flavored' async
functions are ones which only call other maysync functions through their 'maysync context' - that is, they  use the 'a'
methods on maywaitables - for example: `await m_foo().a()` - and the maysync machinery will ultimately call the
appropriate 'leaf' sync or async functions. Being regular python functions they are free to call whatever they like -
for example doing sync IO - but the point is to, ideally, route all IO through maysync functions such that the maysync
code is fully efficiently usable in any context.

Internally, it's not really correct to say that there is 'no event loop' in the maysync context - rather, each
'entrypoint' call to a maysync fn runs within its own tiny event loop.

===

TODO:
 - __del__
 - (test) maysync context managers
 - CancelledError
 - for debug, mask any running eventloop while running maysync code
 - `[CO_ASYNC_GENERATOR] = {k for k, v in dis.COMPILER_FLAG_NAMES.items() if v == 'ASYNC_GENERATOR'}` ? inspect is big..
  - works down to 3.8
 - make_maysync_from_sync can run with asyncio.run_in_thread
 - make_maysync_from_async can run with asyncio.run_soon
"""
import abc
import functools
import inspect
import sys
import threading
import typing as ta


T = ta.TypeVar('T')
T_co = ta.TypeVar('T_co', covariant=True)

O = ta.TypeVar('O')
O_co = ta.TypeVar('O_co', covariant=True)

I = ta.TypeVar('I')
I_contra = ta.TypeVar('I_contra', contravariant=True)

_MaysyncX = ta.TypeVar('_MaysyncX')

_MaysyncRS = ta.TypeVar('_MaysyncRS')
_MaysyncRA = ta.TypeVar('_MaysyncRA')


##


class Maywaitable(ta.Protocol[T_co]):
    """
    The maysync version of `Awaitable[T]`. Non-generator maysync functions return a `Maywaitable`, with the following
    nullary methods:

     - `def s()` - to be called in sync contexts
     - `async def a()` - to be called in async and maysync contexts

    Only the proper method should be called in each context.
    """

    def s(self) -> T_co:
        ...

    def a(self) -> ta.Awaitable[T_co]:
        ...


class MaysyncGenerator(ta.Protocol[O_co, I_contra]):
    """
    The maysync version of `AsyncGenerator[O, I]`. Generator maysync functions return a `MaysyncGenerator`, with the
    following methods:

     - `def s()` - to be called in sync contexts
     - `async def a()` - to be called in async and maysync contexts

    Only the proper method should be called in each context.
    """

    def s(self) -> ta.Generator[O_co, I_contra, None]:
        ...

    def a(self) -> ta.AsyncGenerator[O_co, I_contra]:
        ...


# The maysync equivalent of an async function
MaysyncFn = ta.Callable[..., Maywaitable[T]]  # ta.TypeAlias  # omlish-amalg-typing-no-move

# The maysync equivalent of an async generator function
MaysyncGeneratorFn = ta.Callable[..., MaysyncGenerator[O, I]]  # ta.TypeAlias  # omlish-amalg-typing-no-move


class Maysync_(abc.ABC):  # noqa
    """
    Abstract base class for maysync objects - either MaysyncFn's or MaysyncGeneratorFn's.

    The concrete implementations are module-level implementation detail, and in general users should make a point to
    only interact with the protocols defined above, but introspection can be necessary at times.
    """

    def __init_subclass__(cls, **kwargs):
        if Maysync_ in cls.__bases__ and abc.ABC not in cls.__bases__:
            raise TypeError(cls)

        super().__init_subclass__(**kwargs)

    class FnPair(ta.NamedTuple):
        s: ta.Callable[..., ta.Any]
        a: ta.Callable[..., ta.Any]

    @abc.abstractmethod
    def fn_pair(self) -> ta.Optional[FnPair]:
        """If this maysync object is backed by a (sync, async) pair of functions, returns the pair."""

        raise NotImplementedError

    @abc.abstractmethod
    def cast(self):
        pass

    @abc.abstractmethod
    def __call__(self, *args, **kwargs):
        raise NotImplementedError


class MaysyncFn_(Maysync_, abc.ABC, ta.Generic[T]):  # noqa
    @ta.final
    def cast(self) -> MaysyncFn[T]:
        return ta.cast('MaysyncFn[T]', self)


class MaysyncGeneratorFn_(Maysync_, abc.ABC, ta.Generic[O, I]):  # noqa
    @ta.final
    def cast(self) -> MaysyncGenerator[O, I]:
        return ta.cast('MaysyncGenerator[O, I]', self)


##


class _MaysyncThreadLocal(threading.local):
    context: ta.Optional['_MaysyncContext'] = None


class _MaysyncContext(abc.ABC):
    mode: ta.ClassVar[ta.Literal['s', 'a']]

    @abc.abstractmethod
    def run(self, fn: ta.Callable[..., T], *args: ta.Any, **kwargs: ta.Any) -> T:
        raise NotImplementedError


@ta.final
class _SyncMaysyncContext(_MaysyncContext):
    mode: ta.ClassVar[ta.Literal['s']] = 's'

    def run(self, fn: ta.Callable[..., T], *args: ta.Any, **kwargs: ta.Any) -> T:
        prev = _MaysyncThreadLocal.context
        _MaysyncThreadLocal.context = self

        ph: ta.Any = sys.get_asyncgen_hooks()
        if ph.firstiter is not None or ph.finalizer is not None:
            sys.set_asyncgen_hooks(firstiter=None, finalizer=None)
        else:
            ph = None

        try:
            return fn(*args, **kwargs)

        finally:
            if ph is not None:
                sys.set_asyncgen_hooks(*ph)

            _MaysyncThreadLocal.context = prev


@ta.final
class _AsyncMaysyncContext(_MaysyncContext):
    mode: ta.ClassVar[ta.Literal['a']] = 'a'

    def run(self, fn: ta.Callable[..., T], *args: ta.Any, **kwargs: ta.Any) -> T:
        prev = _MaysyncThreadLocal.context
        _MaysyncThreadLocal.context = self

        try:
            return fn(*args, **kwargs)

        finally:
            _MaysyncThreadLocal.context = prev


##


class _MaywaitableLike(
    abc.ABC,
    ta.Generic[_MaysyncX],
):
    """Abstract base class for the maysync versions of `Awaitable[T]` and `AsyncGenerator[O, I]`."""

    @ta.final
    def __init__(
            self,
            x: _MaysyncX,
            args: ta.Tuple[ta.Any, ...],
            kwargs: ta.Mapping[str, ta.Any],
    ) -> None:
        self._x, self._args, self._kwargs = x, args, kwargs

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}@{id(self):x}({self._x!r})'


class _Maywaitable(
    _MaywaitableLike[_MaysyncX],
    abc.ABC,
    ta.Generic[_MaysyncX, T],
):
    pass


class _MaysyncGenerator(
    _MaywaitableLike[_MaysyncX],
    abc.ABC,
    ta.Generic[_MaysyncX, O, I],
):
    pass


##


class _FpMaysyncFnLike(
    abc.ABC,
    ta.Generic[_MaysyncRS, _MaysyncRA],
):
    """A maysync object backed by an underlying (sync, async) function pair."""

    def __init__(
            self,
            s: ta.Callable[..., _MaysyncRS],
            a: ta.Callable[..., _MaysyncRA],
    ) -> None:
        if s is None:
            raise TypeError(s)
        if a is None:
            raise TypeError(a)
        self._s = s
        self._a = a

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}@{id(self):x}({self._s!r}, {self._a!r})'

    def fn_pair(self) -> ta.Optional[Maysync_.FnPair]:
        return Maysync_.FnPair(
            self._s,
            self._a,
        )

    def __get__(self, instance, owner=None):
        return self.__class__(
            self._s.__get__(instance, owner),  # noqa
            self._a.__get__(instance, owner),  # noqa
        )


@ta.final
class _FpMaysyncFn(
    _FpMaysyncFnLike[T, ta.Awaitable[T]],
    MaysyncFn_[T],
    ta.Generic[T],
):
    def __call__(self, *args, **kwargs):
        return _FpMaywaitable(self, args, kwargs)


@ta.final
class _FpMaywaitable(
    _Maywaitable[_FpMaysyncFn[T], T],
):
    def s(self) -> T:
        return self._x._s(*self._args, **self._kwargs)  # noqa

    def a(self) -> ta.Awaitable[T]:
        if _MaysyncThreadLocal.context is not None:
            return _MaysyncFuture(self)

        return self._x._a(*self._args, **self._kwargs)  # noqa


def make_maysync_fn(
        s: ta.Callable[..., T],
        a: ta.Callable[..., ta.Awaitable[T]],
) -> MaysyncFn[T]:
    """Constructs a MaysyncFn from a (sync, async) function pair."""

    return _FpMaysyncFn(s, a)


@ta.final
class _FpMaysyncGeneratorFn(
    _FpMaysyncFnLike[ta.Generator[O, I, None], ta.AsyncGenerator[O, I]],
    MaysyncGeneratorFn_[O, I],
    ta.Generic[O, I],
):
    def __call__(self, *args, **kwargs):
        return _FpMaysyncGenerator(self, args, kwargs)


@ta.final
class _FpMaysyncGenerator(
    _MaysyncGenerator[_FpMaysyncGeneratorFn[O, I], O, I],
):
    def s(self) -> ta.Generator[O, I, None]:
        return self._x._s(*self._args, **self._kwargs)  # noqa

    def a(self) -> ta.AsyncGenerator[O, I]:
        if (ctx := _MaysyncThreadLocal.context) is not None and ctx.mode == 's':
            async def inner():
                g = self._x._s(*self._args, **self._kwargs)  # noqa

                i: ta.Any = None
                e: ta.Any = None

                while True:
                    try:
                        if e is not None:
                            o = g.throw(e)
                        else:
                            o = g.send(i)
                    except StopIteration as ex:
                        if ex.value is not None:
                            raise TypeError(ex) from None
                        return

                    i = None
                    e = None

                    try:
                        i = yield o
                    except StopIteration as ex:  # noqa
                        raise NotImplementedError  # noqa
                    except StopAsyncIteration as ex:  # noqa
                        raise NotImplementedError  # noqa
                    except BaseException as ex:  # noqa
                        e = ex

            return inner()

        return self._x._a(*self._args, **self._kwargs)  # noqa


def make_maysync_generator_fn(
        s: ta.Callable[..., ta.Generator[O, I, None]],
        a: ta.Callable[..., ta.AsyncGenerator[O, I]],
) -> MaysyncGeneratorFn[O, I]:
    """Constructs a MaysyncGeneratorFn from a (sync, async) generator function pair."""

    return _FpMaysyncGeneratorFn(s, a)


@ta.overload
def make_maysync(
        s: ta.Callable[..., T],
        a: ta.Callable[..., ta.Awaitable[T]],
) -> MaysyncFn[T]:
    ...


@ta.overload
def make_maysync(
        s: ta.Callable[..., ta.Generator[O, I, None]],
        a: ta.Callable[..., ta.AsyncGenerator[O, I]],
) -> MaysyncGeneratorFn[O, I]:
    ...


def make_maysync(s, a):
    """
    Constructs a MaysyncFn or MaysyncGeneratorFn from a (sync, async) function pair, using `inspect.isasyncgenfunction`
    to determine the type.
    """

    if inspect.isasyncgenfunction(a):
        return make_maysync_generator_fn(s, a)
    else:
        return make_maysync_fn(s, a)


##


class _MgMaysyncFnLike(
    abc.ABC,
    ta.Generic[T],
):
    def __init__(
            self,
            mg: ta.Callable[..., T],
    ) -> None:
        self._mg = mg

        functools.update_wrapper(self, mg, updated=())

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}@{id(self):x}({self._mg!r})'

    def fn_pair(self) -> ta.Optional[Maysync_.FnPair]:
        return None

    def __get__(self, instance, owner=None):
        return self.__class__(
            self._mg.__get__(instance, owner),  # noqa
        )

    @abc.abstractmethod  # noqa
    def __call__(self, *args, **kwargs):
        raise NotImplementedError


@ta.final
class _MgMaysyncFn(
    _MgMaysyncFnLike[ta.Awaitable[T]],
    MaysyncFn_[T],
    ta.Generic[T],
):
    def __call__(self, *args, **kwargs):
        return _MgMaywaitable(self, args, kwargs)


@ta.final
class _MgMaysyncDriver:
    def __init__(self, ctx: _MaysyncContext, mg: ta.Any) -> None:
        self.ctx = ctx
        self.mg = mg

    value: ta.Any

    def __iter__(self) -> ta.Generator['_MaysyncFuture', None, None]:
        try:
            a = self.mg.__await__()
            try:
                g = iter(a)
                try:
                    while True:
                        try:
                            f = self.ctx.run(g.send, None)
                        except StopIteration as ex:
                            self.value = ex.value
                            return

                        if not isinstance(f, _MaysyncFuture):
                            raise TypeError(f)

                        yield f
                        del f

                finally:
                    if g is not a:
                        self.ctx.run(g.close)

            finally:
                self.ctx.run(a.close)

        finally:
            self.ctx.run(self.mg.close)


@ta.final
class _MgMaywaitable(
    _Maywaitable[_MgMaysyncFn[T], T],
):
    def _driver(self, ctx: _MaysyncContext) -> _MgMaysyncDriver:
        return _MgMaysyncDriver(ctx, self._x._mg(*self._args, **self._kwargs))  # noqa

    def s(self) -> T:
        for f in (drv := self._driver(_SyncMaysyncContext())):
            f.s()
            del f

        return drv.value

    def a(self) -> ta.Awaitable[T]:
        if (ctx := _MaysyncThreadLocal.context) is None or ctx.mode == 'a':
            return self._x._mg(*self._args, **self._kwargs)  # noqa

        async def inner():
            for f in (drv := self._driver(_AsyncMaysyncContext())):
                await f.a()
                del f

            return drv.value

        return inner()


@ta.final
class _MgMaysyncGeneratorFn(
    _MgMaysyncFnLike[ta.AsyncGenerator[O, I]],
    MaysyncGeneratorFn_[O, I],
    ta.Generic[O, I],
):
    def __call__(self, *args, **kwargs):
        return _MgMaysyncGenerator(self, args, kwargs)


@ta.final
class _MgMaysyncGeneratorDriver:
    def __init__(self, ctx: _MaysyncContext, ag: ta.Any) -> None:
        self.ctx = ctx
        self.ag = ag

    def __iter__(self) -> ta.Generator[
        ta.Union[
            ta.Tuple[ta.Literal['f'], '_MaysyncFuture'],
            ta.Tuple[ta.Literal['o'], ta.Any],
        ],
        ta.Union[
            ta.Tuple[ta.Any, BaseException],
            None,
        ],
        None,
    ]:
        try:
            ai = self.ag.__aiter__()
            try:
                i: ta.Any = None
                e: ta.Any = None

                while True:
                    if e is not None:
                        coro = ai.athrow(e)
                    else:
                        coro = ai.asend(i)

                    i = None
                    e = None

                    for f in (drv := _MgMaysyncDriver(self.ctx, coro)):
                        if (x := (yield ('f', f))) is not None:
                            raise RuntimeError(x)

                        del f

                    i, e = yield ('o', drv.value)  # type: ignore[misc]

            finally:
                if ai is not self.ag:
                    for f in _MgMaysyncDriver(self.ctx, ai.aclose()):
                        yield ('f', f)

        finally:
            for f in _MgMaysyncDriver(self.ctx, self.ag.aclose()):
                yield ('f', f)


@ta.final
class _MgMaysyncGenerator(
    _MaysyncGenerator[_MgMaysyncGeneratorFn[O, I], O, I],
):
    def _driver(self, ctx: _MaysyncContext) -> _MgMaysyncGeneratorDriver:
        return _MgMaysyncGeneratorDriver(ctx, self._x._mg(*self._args, **self._kwargs))  # noqa

    def s(self) -> ta.Generator[O, I, None]:
        di = iter(self._driver(_SyncMaysyncContext()))

        ie: ta.Any = None

        while True:
            try:
                t, x = di.send(ie)
            except StopAsyncIteration:
                return
            except StopIteration:
                raise RuntimeError from None

            ie = None

            if t == 'f':
                x.s()

            elif t == 'o':
                try:
                    ie = ((yield x), None)  # type: ignore[misc]
                except BaseException as ex:  # noqa
                    ie = (None, ex)

            else:
                raise RuntimeError((t, x))

            del x

    def a(self) -> ta.AsyncGenerator[O, I]:
        if _MaysyncThreadLocal.context is not None:
            return self._x._mg(*self._args, **self._kwargs)  # noqa

        async def inner():
            di = iter(self._driver(_AsyncMaysyncContext()))

            ie: ta.Any = None

            while True:
                try:
                    t, x = di.send(ie)
                except StopAsyncIteration:
                    return
                except StopIteration:
                    raise RuntimeError from None

                ie = None

                if t == 'f':
                    await x.a()

                elif t == 'o':
                    try:
                        ie = ((yield x), None)
                    except BaseException as ex:  # noqa
                        ie = (None, ex)

                else:
                    raise RuntimeError((t, x))

                del x

        return inner()


def maysync_fn(
        m: ta.Callable[..., ta.Awaitable[T]],
) -> MaysyncFn[T]:
    """Constructs a MaysyncFn from a 'maysync flavored' async function."""

    return _MgMaysyncFn(m)


def maysync_generator_fn(
        m: ta.Callable[..., ta.AsyncGenerator[O, I]],
) -> MaysyncGeneratorFn[O, I]:
    """Constructs a MaysyncGeneratorFn from a 'maysync flavored' async generator function."""

    return _MgMaysyncGeneratorFn(m)


@ta.overload
def maysync(
        m: ta.Callable[..., ta.Awaitable[T]],
) -> MaysyncFn[T]:
    ...


@ta.overload
def maysync(
        m: ta.Callable[..., ta.AsyncGenerator[O, I]],
) -> MaysyncGeneratorFn[O, I]:
    ...


def maysync(m):
    """
    Constructs a MaysyncFn or MaysyncGeneratorFn from 'maysync flavored' async function or async generator function,
    using `inspect.isasyncgenfunction` to determine the type. Usable as a decorator.
    """

    if inspect.isasyncgenfunction(m):
        return maysync_generator_fn(m)
    else:
        return maysync_fn(m)


##


@ta.final
class _MaysyncFutureNotAwaitedError(RuntimeError):
    pass


@ta.final
class _MaysyncFuture(ta.Generic[T]):
    def __init__(
            self,
            x: Maywaitable[T],
    ) -> None:
        self._x = x

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}@{id(self):x}({self._x!r}, done={self.done!r})'

    done: bool = False
    result: T
    error: ta.Optional[BaseException] = None

    def __await__(self):
        if not self.done:
            yield self
        if not self.done:
            raise _MaysyncFutureNotAwaitedError
        if self.error is not None:
            raise self.error
        else:
            return self.result

    def s(self) -> None:
        if self.done:
            return

        try:
            self.result = self._x.s()
        except BaseException as ex:  # noqa
            self.error = ex
        self.done = True

    async def a(self) -> None:
        if self.done:
            return

        try:
            self.result = await self._x.a()
        except BaseException as ex:  # noqa
            self.error = ex
        self.done = True



import dataclasses as dc
import importlib.resources
import json
import typing as ta


##


PACKAGE_CONFIG_FILE_NAME = '.dataclasses.json'


@dc.dataclass(frozen=True, kw_only=True)
class PackageConfig:
    codegen: bool = False

    def __init_subclass__(cls, **kwargs):
        raise TypeError

    @classmethod
    def loads(cls, s: str) -> 'PackageConfig':
        return cls(**json.loads(s))

    def dumps(self) -> str:
        return json.dumps(dc.asdict(self), indent=2)


DEFAULT_PACKAGE_CONFIG = PackageConfig()


##


class PackageConfigCache:
    def __init__(self) -> None:
        super().__init__()

        self._dct: dict[str, PackageConfig | None] = {}

    @ta.overload
    def get(self, pkg: str, default: PackageConfig) -> PackageConfig:
        ...

    @ta.overload
    def get(self, pkg: str, default: PackageConfig | None = None) -> PackageConfig | None:
        ...

    def get(self, pkg, default=None):
        try:
            return self._dct[pkg]
        except KeyError:
            pass

        try:
            s = importlib.resources.read_text(pkg, PACKAGE_CONFIG_FILE_NAME)
        except FileNotFoundError:
            self._dct[pkg] = None
        else:
            c = PackageConfig.loads(s)
            self._dct[pkg] = c
            return c

        if '.' not in pkg:
            return default

        return self.get(pkg.rpartition('.')[0], default)


PACKAGE_CONFIG_CACHE = PackageConfigCache()



import importlib.util
import types
import typing as ta


##


def lazy_import(
        name: str,
        package: str | None = None,
        *,
        optional: bool = False,
        cache_failure: bool = False,
) -> ta.Callable[[], ta.Any]:
    result = not_set = object()

    def inner():
        nonlocal result

        if result is not not_set:
            if isinstance(result, Exception):
                raise result
            return result

        try:
            mod = importlib.import_module(name, package=package)

        except Exception as e:
            if optional:
                if cache_failure:
                    result = None
                return None

            if cache_failure:
                result = e
            raise

        result = mod
        return mod

    return inner


def proxy_import(
        name: str,
        package: str | None = None,
        extras: ta.Iterable[str] | None = None,
) -> types.ModuleType:
    if isinstance(extras, str):
        raise TypeError(extras)

    omod = None

    def __getattr__(att):  # noqa
        nonlocal omod
        if omod is None:
            omod = importlib.import_module(name, package=package)
            if extras:
                for x in extras:
                    importlib.import_module(f'{name}.{x}', package=package)
        return getattr(omod, att)

    lmod = types.ModuleType(name)
    lmod.__getattr__ = __getattr__  # type: ignore[method-assign]
    return lmod
